﻿unit USortProjects;

interface

{$IF RTLVersion >= 23.00}
  {$DEFINE USENAMESPACE} // XE2 or later
{$IFEND}

procedure Register;

implementation

uses
{$IFDEF USENAMESPACE}
  Winapi.Windows, System.SysUtils, System.Classes, System.Generics.Collections,
  System.Generics.Defaults, System.Rtti, System.TypInfo, Vcl.Menus, ToolsAPI;
{$ELSE}
  Windows, SysUtils, Classes, Generics.Collections, Generics.Defaults,
  Rtti, TypInfo, Menus, ToolsAPI;
{$ENDIF}

type
  TProjectMenuCreatorNotifier = class(TNotifierObject, INTAProjectMenuCreatorNotifier)
  private
    function AddMenu(const Ident: string): TMenuItem;
    function CanHandle(const Ident: string): Boolean;
    procedure MenuClick(Sender: TObject);
  end;

const
  {$IFDEF VER210}coreide = 'coreide140.bpl';{$ENDIF} // 2010
  {$IFDEF VER220}coreide = 'coreide150.bpl';{$ENDIF} // XE
  {$IFDEF VER230}coreide = 'coreide160.bpl';{$ENDIF} // XE2
  {$IFDEF VER240}coreide = 'coreide170.bpl';{$ENDIF} // XE3
  {$IFDEF VER250}coreide = 'coreide180.bpl';{$ENDIF} // XE4
  {$IFDEF VER260}coreide = 'coreide190.bpl';{$ENDIF} // XE5
  {$IFDEF VER270}coreide = 'coreide200.bpl';{$ENDIF} // XE6
  {$IFDEF VER280}coreide = 'coreide210.bpl';{$ENDIF} // XE7
  {$IFDEF VER290}coreide = 'coreide220.bpl';{$ENDIF} // XE8
  {$IFDEF VER300}coreide = 'coreide230.bpl';{$ENDIF} // 10 Seattle
  {$IFDEF VER310}coreide = 'coreide240.bpl';{$ENDIF} // 10.1 Berlin
  {$IFDEF VER320}coreide = 'coreide250.bpl';{$ENDIF} // 10.2 Tokyo

  sSortProjectsJA = 'プロジェクトをソート(&O)';
  sSortAscendingJA = '昇順(&A)';
  sSortDescendingJA = '降順(&D)';
  sSortProjectsEN = 'S&ort Projects';
  sSortAscendingEN = '&Ascending';
  sSortDescendingEN = '&Descending';

procedure TProjectGroup_StructureChanged(Self: TObject); external coreide name '@Projectgroup@TProjectGroup@StructureChanged$qqrv';

{ TProjectMenuCreatorNotifier }

function TProjectMenuCreatorNotifier.AddMenu(const Ident: string): TMenuItem;
begin
  Result := TMenuItem.Create(nil);
  Result.Add(TMenuItem.Create(Result));
  Result.Add(TMenuItem.Create(Result));

  if SameText(PreferredUILanguages, 'JA') then
  begin
    Result.Caption := sSortProjectsJA;
    Result.Items[0].Caption := sSortAscendingJA;
    Result.Items[1].Caption := sSortDescendingJA;
  end
  else begin
    Result.Caption := sSortProjectsEN;
    Result.Items[0].Caption := sSortAscendingEN;
    Result.Items[1].Caption := sSortDescendingEN;
  end;

  Result.Items[0].OnClick := MenuClick;
  Result.Items[1].OnClick := MenuClick;
end;

function TProjectMenuCreatorNotifier.CanHandle(const Ident: string): Boolean;
begin
  Result := SameText(Ident, 'ProjectGroupContainer');
end;

procedure TProjectMenuCreatorNotifier.MenuClick(Sender: TObject);
var
  isAscending: Boolean;
  objCodeIProjectGroup, objProjectGroup, objStdProjectContainer: TObject;
  ctx: TRttiContext;
  typ: TRttiType;
  fld: TRttiField;
  methModified: TRttiMethod;
  prop: TRttiProperty;
  info: PPropInfo;
  intf: IInterface;
  list: IInterfaceList;
  i: Integer;
  arr: array of TPair<string,IInterface>;
begin
  isAscending := TMenuItem(Sender).MenuIndex = 0;

  objCodeIProjectGroup := (BorlandIDEServices as IOTAModuleServices).MainProjectGroup as TObject;
  typ := ctx.GetType(objCodeIProjectGroup.ClassType);
  if typ = nil then Exit;
  fld := typ.GetField('FProjectGroup');
  if fld = nil then Exit;
  objProjectGroup := fld.GetValue(objCodeIProjectGroup).AsObject;
  if objProjectGroup = nil then Exit;

  typ := ctx.GetType(objProjectGroup.ClassType);
  if typ = nil then Exit;
  methModified := typ.GetMethod('Modified');
  if methModified = nil then Exit;
  fld := typ.GetField('FItems');
  if fld = nil then Exit;
  intf := fld.GetValue(objProjectGroup).AsInterface;
  if intf = nil then Exit;

  list := intf as IInterfaceList;
  if list.Count <= 1 then Exit;
  objStdProjectContainer := list[0] as TObject;
  typ := ctx.GetType(objStdProjectContainer.ClassType);
  if typ = nil then Exit;
  prop := typ.GetProperty('FileName');
  if prop = nil then Exit;
  info := TRttiInstanceProperty(prop).PropInfo;
  if info = nil then Exit;

  SetLength(arr, list.Count);
  for i := 0 to list.Count-1 do
  begin
    arr[i].Key := ExtractFileName(GetStrProp(list[i] as TObject, info));
    arr[i].Value := list[i];
  end;

  TArray.Sort<TPair<string,IInterface>>(arr, TComparer<TPair<string,IInterface>>.Construct(
    function(const Left, Right: TPair<string,IInterface>): Integer
    begin
      if isAscending then
        Result := CompareText(Left.Key, Right.Key)
      else
        Result := CompareText(Right.Key, Left.Key);
    end));

  list.Clear;
  for i := Low(arr) to High(arr) do
    list.Add(arr[i].Value);
  TProjectGroup_StructureChanged(objProjectGroup);
  methModified.Invoke(objProjectGroup, []);
end;

var
  MenuNotifierIndex: Integer = -1;

procedure Register;
begin
  MenuNotifierIndex := (BorlandIDEServices as IOTAProjectManager).AddMenuCreatorNotifier(TProjectMenuCreatorNotifier.Create);
end;

procedure Unregister;
begin
  if MenuNotifierIndex >= 0 then
    (BorlandIDEServices as IOTAProjectManager).RemoveMenuCreatorNotifier(MenuNotifierIndex);
end;

initialization
finalization
  Unregister;
end.